# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
from influxdb import InfluxDBClient

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

log = logging.getLogger(__name__)


class InfluxDBStorage:

    capabilities = ['readings']

    def __init__(self, dsn, measurement='ldi_readings', empty_tables=False, dry_run=False):

        self.dry_run = dry_run
        self.measurement = measurement

        self.buffer = []
        self.is_udp = False

        dsn_parts = urlparse(dsn)
        if dsn_parts.scheme.startswith('udp+'):
            self.is_udp = True
            self.db = InfluxDBClient.from_dsn(dsn, udp_port=dsn_parts.port, timeout=5)
        else:
            self.db = InfluxDBClient.from_dsn(dsn, timeout=5)
            self.db.create_database(self.db._database)


    def emit(self, reading):
        """
        Will store these kind of records to InfluxDB:
        {
            "measurement": "ldi_readings",
            "tags": {
                "station_id": 28,
                "sensor_id": 658
            },
            "time": "2018-12-12T02:21:08Z",
            "fields": {
                "P1": 13.35,
                "P2": 8.05
            }
        }
        """

        # Debugging
        #import json; print(json.dumps(reading, indent=4))

        # Copy measurement values and delete "time" field.
        data = dict(reading.data)
        del data['time']

        # Build InfluxDB record.
        record = {
            "measurement": self.measurement,
            "tags": {
                "station_id": reading.station.station_id,
                "sensor_id": reading.station.sensor_id,
            },
            "time": reading.data.time,
            "fields": data
        }

        if 'position' in reading.station and 'geohash' in reading.station.position:
            record['tags']['geohash'] = reading.station.position.geohash

        # Store into buffer.
        self.buffer.append(record)

        if self.is_udp and len(self.buffer) > 50:
            self.flush()

    def flush(self, final=False):
        # Store into database.
        #print('Writing points:', len(self.buffer))
        self.db.write_points(self.buffer)
        self.buffer = []



# -*- coding: utf-8 -*-
# (c) 2017,2018 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017,2018 Richard Pobering <richard@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import os
import sys
import glob
import logging
import traceback

from six import StringIO


def setup_logging(level=logging.INFO):
    log_format = '%(asctime)-15s [%(name)-30s] %(levelname)-7s: %(message)s'
    logging.basicConfig(
        format=log_format,
        stream=sys.stderr,
        level=level)

    # TODO: Control debug logging of HTTP requests through yet another commandline option "--debug-http" or "--debug-requests"
    requests_log = logging.getLogger('requests')
    requests_log.setLevel(logging.WARN)


def normalize_options(options):
    normalized = {}
    for key, value in options.items():
        key = key.strip('--<>')
        normalized[key] = value
    return normalized


def read_list(data, separator=u','):
    if data is None:
        return []
    result = list(map(lambda x: x.strip(), data.split(separator)))
    if len(result) == 1 and not result[0]:
        result = []
    return result


def exception_traceback(exc_info=None):
    """
    Return a string containing a traceback message for the given
    exc_info tuple (as returned by sys.exc_info()).

    from setuptools.tests.doctest
    """

    if not exc_info:
        exc_info = sys.exc_info()

    # Get a traceback message.
    excout = StringIO()
    exc_type, exc_val, exc_tb = exc_info
    traceback.print_exception(exc_type, exc_val, exc_tb, file=excout)
    return excout.getvalue()


def to_list(obj):
    """Convert an object to a list if it is not already one"""
    if not isinstance(obj, (list, tuple)):
        obj = [obj, ]
    return obj


def invalidate_dogpile_cache(cache_instance, hard=True):
    """
    The default invalidation system works by setting
    a current timestamp (using ``time.time()``)
    representing the "minimum creation time" for
    a value.  Any retrieved value whose creation
    time is prior to this timestamp
    is considered to be stale.  It does not
    affect the data in the cache in any way, and is
    **local to this instance of :class:`.CacheRegion`.**

    .. warning::

        The :meth:`.CacheRegion.invalidate` method's default mode of
        operation is to set a timestamp **local to this CacheRegion
        in this Python process only**.   It does not impact other Python
        processes or regions as the timestamp is **only stored locally in
        memory**.  To implement invalidation where the
        timestamp is stored in the cache or similar so that all Python
        processes can be affected by an invalidation timestamp, implement a
        custom :class:`.RegionInvalidationStrategy`.

    The method supports both "hard" and "soft" invalidation
    options.  With "hard" invalidation,
    :meth:`.CacheRegion.get_or_create` will force an immediate
    regeneration of the value which all getters will wait for.
    With "soft" invalidation, subsequent getters will return the
    "old" value until the new one is available.
    """
    cache_instance.invalidate(hard=hard)


def find_files_glob(pattern):
    return sorted(glob.glob(pattern, recursive=True))


def find_files_walk(path, suffix):
    for root, subdirs, files in os.walk(path):
        # https://python-forum.io/Thread-why-i-don-t-like-os-walk?pid=35718#pid35718
        subdirs[:] = sorted(subdirs)  # sorts the subdirs
        for name in sorted(files):
            realname = os.path.join(root, name)
            if realname.endswith(suffix):
                yield realname

find_files = find_files_glob



# -*- coding: utf-8 -*-
# (c) 2017,2018 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017,2018 Richard Pobering <richard@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import sys
import logging
from docopt import docopt, DocoptExit
from luftdatenpumpe import __appname__, __version__
from luftdatenpumpe.geo import disable_nominatim_cache
from luftdatenpumpe.target import resolve_target_handler
from luftdatenpumpe.util import normalize_options, setup_logging, read_list
from luftdatenpumpe.core import LuftdatenPumpe
from luftdatenpumpe.engine import LuftdatenEngine

log = logging.getLogger(__name__)


def run():
    """
    Usage:
      luftdatenpumpe stations [options] [--target=<target>]...
      luftdatenpumpe readings [options] [--target=<target>]...
      luftdatenpumpe --version
      luftdatenpumpe (-h | --help)

    Options:
      --source=<source>             Data source, either "api" or "file://" [default: api].
      --station=<stations>          Filter data by given location ids, comma-separated.
      --sensor=<sensors>            Filter data by given sensor ids, comma-separated.
      --sensor-type=<sensor-types>  Filter data by given sensor types, comma-separated.
      --reverse-geocode             Compute geographical address using the Nominatim reverse geocoder
      --target=<target>             Data output target
      --create-database-view        Create database view like "ldi_view" spanning all tables.
      --disable-nominatim-cache     Disable Nominatim reverse geocoder cache
      --progress                    Show progress bar
      --version                     Show version information
      --dry-run                     Skip publishing to MQTT bus
      --debug                       Enable debug messages
      -h --help                     Show this screen


    Station list examples:

      # Display metadata for given stations in JSON format
      luftdatenpumpe stations --station=28,1071 --reverse-geocode

      # Display metadata for given sensors in JSON format
      luftdatenpumpe stations --sensor=657,2130 --reverse-geocode

      # Display list of stations in JSON format made of value/text items, suitable for use as a Grafana JSON data source
      luftdatenpumpe stations --station=28,1071 --reverse-geocode --target=json.grafana.vt+stream://sys.stdout

      # Display list of stations in JSON format made of key/name items, suitable for use as a mapping in Grafana Worldmap Panel
      luftdatenpumpe stations --station=28,1071 --reverse-geocode --target=json.grafana.kn+stream://sys.stdout

      # Write list of stations and metadata to RDBMS database (PostgreSQL), also display on STDERR
      luftdatenpumpe stations --station=28,1071 --reverse-geocode --target=postgresql:///weatherbase --target=json+stream://sys.stderr

      # Read station information from RDBMS database (PostgreSQL) and format for Grafana Worldmap Panel
      luftdatenpumpe stations --source=postgresql:///weatherbase --target=json.grafana.kn+stream://sys.stdout


    Live data examples (InfluxDB):

      # Store into InfluxDB running on "localhost"
      luftdatenpumpe readings --station=28,1071 --target=influxdb://localhost:8086/luftdaten_info

      # Store into InfluxDB, with UDP
      luftdatenpumpe readings --station=28,1071 --target=udp+influxdb://localhost:4445/luftdaten_info

      # Store into InfluxDB, with authentication
      luftdatenpumpe readings --station=28,1071 --target=influxdb://username:password@localhost:8086/luftdaten_info


    Archive data examples (InfluxDB):

      # Mirror archive of luftdaten.info
      wget --mirror --continue --no-host-directories --directory-prefix=/var/spool/archive.luftdaten.info http://archive.luftdaten.info/

      # Ingest station information from CSV archive files, store into PostgreSQL
      luftdatenpumpe stations --source=file:///var/spool/archive.luftdaten.info --target=postgresql:///weatherbase --reverse-geocode --progress

      # Ingest readings from CSV archive files, store into InfluxDB
      luftdatenpumpe readings --source=file:///var/spool/archive.luftdaten.info --station=483 --sensor=988 --target=influxdb://localhost:8086/luftdaten_info --progress

      # Ingest most early readings
      luftdatenpumpe readings --source=file:///var/spool/archive.luftdaten.info/2015-10-*

      # Ingest most early PMS sensors
      luftdatenpumpe readings --source=file:///var/spool/archive.luftdaten.info/2017-1*/*pms*.csv


    Live data examples (MQTT):

      # Publish data to topic "luftdaten.info" at MQTT broker running on "localhost"
      luftdatenpumpe readings --station=28,1071 --target=mqtt://localhost/luftdaten.info

      # MQTT publishing, with authentication
      luftdatenpumpe readings --station=28,1071 --target=mqtt://username:password@localhost/luftdaten.info


    Combined examples:

      # Write stations to STDERR and PostgreSQL
      luftdatenpumpe stations --station=28,1071 --target=json+stream://sys.stderr --target=postgresql:///weatherbase

      # Write readings to STDERR, MQTT and InfluxDB
      luftdatenpumpe readings --station=28,1071 --target=json+stream://sys.stderr --target=mqtt://localhost/luftdaten.info --target=influxdb://localhost:8086/luftdaten_info


    """

    # Parse command line arguments
    options = normalize_options(docopt(run.__doc__, version=__appname__ + ' ' + __version__))

    # Setup logging
    debug = options.get('debug')
    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    setup_logging(log_level)

    #log.info('Options: {}'.format(json.dumps(options, indent=4)))

    # Optionally, decode filters by station id and/or sensor id
    filter = {}
    for filter_name in ['station', 'sensor']:
        if options[filter_name]:
            filter[filter_name] = list(map(int, read_list(options[filter_name])))
    for filter_name in ['sensor-type']:
        if options[filter_name]:
            filter[filter_name] = list(map(str.lower, read_list(options[filter_name])))
    log.info('Applying filter: {}'.format(filter))

    # Fake data source. Currently always LDI.
    # TODO: Add more data sources.
    datasource = 'ldi'
    datasource_humanized = datasource.upper()

    # Default output target is STDOUT.
    if not options['target']:
        options['target'] = ['json+stream://sys.stdout']

    # Optionally disable Nominatim cache.
    if options['disable-nominatim-cache']:
        # Invalidate the Nominatim cache; this applies only for this session, it will _not_ _purge_ all data at once.
        disable_nominatim_cache()

    # Create database view and exit.
    if options['create-database-view']:
        log.info('Creating database view')
        for target in options['target']:
            if target.startswith('postgresql:'):
                handler = resolve_target_handler(target)
                handler.create_view()
        sys.exit()

    # The main workhorse.
    pump = LuftdatenPumpe(
        source=options['source'],
        filter=filter,
        reverse_geocode=options['reverse-geocode'],
        progressbar=options['progress'],
        dry_run=options['dry-run'],
    )

    # Acquire data.
    data = None
    kind = None
    if options['stations']:
        kind = 'stations'
        log.info('Acquiring list of stations from {}. source={}'.format(datasource_humanized, options['source']))
        data = pump.get_stations()
        log.info('Acquired #{} stations'.format(len(data)))

    elif options['readings']:
        kind = 'readings'
        log.info('Acquiring readings from {}. source={}'.format(datasource_humanized, options['source']))
        data = pump.get_readings()

    else:
        raise DocoptExit('Subcommand not implemented')

    # Sanity checks.
    if data is None:
        log.error('No data to process')
        sys.exit(2)

    # Create and run output processing engine.
    log.info('Will publish data to {}'.format(options['target']))
    engine = LuftdatenEngine(kind, options['target'], options.get('dry-run', False))
    engine.process(data)



# -*- coding: utf-8 -*-
# (c) 2017,2018 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017,2018 Richard Pobering <richard@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import logging
import dataset
from copy import deepcopy
from collections import OrderedDict

log = logging.getLogger(__name__)


class RDBMSStorage:

    capabilities = ['stations']

    def __init__(self, uri, empty_tables=False, dry_run=False):

        self.dry_run = dry_run

        self.db = dataset.connect(uri)

        # Optionally, empty all tables
        if empty_tables:
            for tablename in self.db.tables:
                if tablename.startswith('ldi_'):
                    self.db.query('DELETE FROM {}'.format(tablename))

        # Create tables.
        self.db.create_table('ldi_stations', primary_id='station_id')
        self.db.create_table('ldi_osmdata', primary_id='station_id')
        self.db.create_table('ldi_sensors', primary_id='sensor_id')

        # Assign table handles.
        self.stationtable = self.db['ldi_stations']
        self.sensorstable = self.db['ldi_sensors']
        self.osmtable = self.db['ldi_osmdata']

    def emit(self, station):
        return self.store_station(station)

    def flush(self, final=False):
        pass

    def store_station(self, station):

        # Debugging
        #log.info(station)

        # Station table
        stationdata = OrderedDict()
        stationdata['station_id'] = station.station_id
        for key, value in station.items():
            if key.startswith('name'):
                stationdata[key] = value
        stationdata.update(station.position)
        self.stationtable.upsert(stationdata, ['station_id'])

        # Sensors table
        for sensor in station['sensors']:
            sensordata = OrderedDict()
            sensordata['station_id'] = station.station_id
            sensordata.update(sensor)
            self.sensorstable.upsert(sensordata, ['sensor_id'])

        # OSM table
        if 'location' in station:
            osmdata = OrderedDict()
            osmdata['station_id'] = station.station_id
            location = deepcopy(station.location)
            for key, value in location.address.items():
                #key = 'address_' + key
                location[key] = value
            del location['address']

            if 'address_more' in location:
                osmdata['address_more'] = json.dumps(location.address_more)
                del location['address_more']

            # TODO: Also store bounding box
            del location['boundingbox']

            osmdata.update(location)

            # Prefix designated column name with 'osm_'.
            osmdata_real = OrderedDict()
            for key, value in osmdata.items():
                if key not in ['station_id']:
                    if not key.startswith('osm_'):
                        key = 'osm_' + key
                osmdata_real[key] = value

            self.osmtable.upsert(osmdata_real, ['station_id'])

    def dump_tables(self):
        for table in [self.stationtable, self.sensorstable, self.osmtable]:
            print('### Table name:', table.name)
            for record in table.all():
                print(self.json_markdown_dumps(record))
                print()

    def demo_query(self):

        #print(dir(self.stationtable.table))

        # https://dataset.readthedocs.io/en/latest/quickstart.html#running-custom-sql-queries
        expression = 'SELECT * FROM ldi_stations, ldi_osmdata WHERE ldi_stations.station_id = ldi_osmdata.station_id'
        print('### Expression:', expression)
        result = self.db.query(expression)
        for record in result:
            print(self.json_markdown_dumps(record))

    @staticmethod
    def json_markdown_dumps(thing):
        tpl = "```\n{}\n```"
        content = json.dumps(thing, indent=4)
        return tpl.format(content)

    @classmethod
    def get_dialects(cls):
        results = []
        try:
            from sqlalchemy.dialects import __all__ as sql_dialects
            results = list(sql_dialects)
        except:
            pass
        #log.info('SQLAlchemy dialects: %s', results)
        return results

    def create_view(self):

        # Inquire osmdata fields.
        osmdata = self.db.query("""
            SELECT concat(TABLE_NAME, '.', COLUMN_NAME) AS name
            FROM INFORMATION_SCHEMA.COLUMNS
            WHERE TABLE_NAME='ldi_osmdata' AND COLUMN_NAME NOT IN ('station_id')
            ORDER BY ORDINAL_POSITION
        """)
        osmdata_columns = []
        for record in osmdata:
            osmdata_columns.append(record['name'])

        # Create unified view.
        osmdata_columns_expression = ', '.join(osmdata_columns)
        view = """
        DROP VIEW IF EXISTS ldi_network;
        CREATE VIEW ldi_network AS
            SELECT
              ldi_stations.station_id,
              ldi_stations.name, ldi_stations.latitude, ldi_stations.longitude, ldi_stations.altitude, ldi_stations.country, ldi_stations.geohash,
              ldi_sensors.sensor_id, ldi_sensors.sensor_type,
              concat(ldi_osmdata.osm_state, ' » ', ldi_osmdata.osm_city) AS state_and_city,
              concat(ldi_stations.name, ' (#', CAST(ldi_stations.station_id AS text), ')') AS name_and_id,
              concat(ldi_osmdata.osm_country, ' (', ldi_osmdata.osm_country_code, ')') AS country_and_countrycode,
              concat(concat_ws(', ', ldi_osmdata.osm_state, ldi_osmdata.osm_country), ' (', ldi_osmdata.osm_country_code, ')') AS state_and_country,
              concat(concat_ws(', ', ldi_osmdata.osm_city, ldi_osmdata.osm_state, ldi_osmdata.osm_country), ' (', ldi_osmdata.osm_country_code, ')') AS city_and_state_and_country,
              {}
            FROM
              ldi_stations, ldi_osmdata, ldi_sensors
            WHERE
              ldi_stations.station_id = ldi_osmdata.station_id AND
              ldi_stations.station_id = ldi_sensors.station_id
            ORDER BY
              osm_country_code, state_and_city, name_and_id, sensor_type
        """.format(osmdata_columns_expression)
        self.db.query(view)



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
from luftdatenpumpe.target import resolve_target_handler

log = logging.getLogger(__name__)


class LuftdatenEngine:

    def __init__(self, kind, targets, dry_run, batch_size=2500):
        self.kind = kind
        self.targets = targets
        self.dry_run = dry_run
        self.batch_size = batch_size

    def process(self, data):

        # Configure target subsystems.
        targets = []
        for target_expression in self.targets:
            log.info('Configuring target {}'.format(target_expression))
            target = resolve_target_handler(target_expression, dry_run=self.dry_run)

            if target is None:
                log.error('Could not resolve target {}'.format(target_expression))
                continue

            if self.kind not in target.capabilities:
                log.warning('Target {} does not handle kind "{}"'.format(target_expression, self.kind))
                continue

            targets.append(target)

        # Emit to active target subsystems.
        log.info('Emitting data to target subsystems, this might take some time')
        item_count = 0
        for item in data:
            for target in targets:
                target.emit(item)
            item_count += 1

            # Preliminary flush each $batch_size items.
            if item_count % self.batch_size == 0:
                target.flush()

        # Signal readyness to each target subsystem.
        for target in targets:
            target.flush(final=True)

        log.info('Processed {} records'.format(item_count))



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import types
import logging

from munch import Munch
from urllib.parse import urlparse

from luftdatenpumpe.target.influxdb import InfluxDBStorage
from luftdatenpumpe.target.stream import StreamTarget
from luftdatenpumpe.target.rdbms import RDBMSStorage
from luftdatenpumpe.target.mqtt import MQTTAdapter

log = logging.getLogger(__name__)


def json_formatter(data):
    if isinstance(data, types.GeneratorType):
        data = list(data)
    return json.dumps(data, indent=4)


def json_grafana_formatter_base(stations):
    entries = []
    for station in stations:
        if 'name' in station:
            station_name = station.name
        else:
            station_name = u'Station #{}, {}'.format(station.station_id, station.position.country)
        entry = {'value': station.station_id, 'text': station_name}
        entries.append(entry)
    return entries


def json_grafana_formatter_vt(stations):
    """
    Format list of stations in JSON format made of value/text items,
    suitable for use as a Grafana JSON data source.

    Example::

        luftdatenpumpe stations --source=postgresql:///weatherbase --target=json.grafana.vt+stream://sys.stdout

    Emits items like::

        {
            "value": 22,
            "text": "Steglen, Haslach, Herrenberg, Vereinbarte  der Stadt Herrenberg, Baden-W\u00fcrttemberg, DE"
        }

    """
    entries = json_grafana_formatter_base(stations)
    return json_formatter(entries)


def json_grafana_formatter_kn(stations):
    """
    Format list of stations in JSON format made of key/name items,
    suitable for use as a mapping in Grafana Worldmap Panel.

    Example::

        luftdatenpumpe stations --source=postgresql:///weatherbase --target=json.grafana.kn+stream://sys.stdout

    Emits items like::

        {
            "key": "22",
            "name": "Steglen, Haslach, Herrenberg, Vereinbarte  der Stadt Herrenberg, Baden-W\u00fcrttemberg, DE"
        }

    """
    entries = []
    for station in json_grafana_formatter_base(stations):
        entry = {'key': str(station['value']), 'name': station['text']}
        entries.append(entry)
    return json_formatter(entries)


def resolve_target_handler(target, dry_run=False):
    handler = None

    url = Munch(urlparse(target)._asdict())
    log.debug('Resolving target: %s', json.dumps(url))

    formatter = lambda x: x
    if '+' in url.scheme:
        format, scheme = url.scheme.split('+')
        url.scheme = scheme
        if format.startswith('json.grafana.vt'):
            formatter = json_grafana_formatter_vt
        elif format.startswith('json.grafana.kn'):
            formatter = json_grafana_formatter_kn
        elif format.startswith('json'):
            formatter = json_formatter
        formatter.format = format

    #effective_url = urlunparse(url.values())

    if url.scheme == 'stream':

        # FIXME: There might be dragons?
        import sys
        stream = eval(url.netloc)

        handler = StreamTarget(stream, formatter)

    elif url.scheme in RDBMSStorage.get_dialects():
        handler = RDBMSStorage(target, dry_run=dry_run)

    elif url.scheme == 'mqtt':
        handler = MQTTAdapter(target, dry_run=dry_run)

    elif url.scheme == 'influxdb':
        handler = InfluxDBStorage(target, dry_run=dry_run)

    return handler



# -*- coding: utf-8 -*-
# (c) 2017,2018 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017,2018 Richard Pobering <richard@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import re
import sys
import json
import redis
import logging

from tqdm import tqdm
from munch import Munch, munchify
from tablib import Dataset
from requests_cache import CachedSession
from luftdatenpumpe.geo import geohash_encode, resolve_location, improve_location, format_address
from luftdatenpumpe.target import RDBMSStorage
from luftdatenpumpe.util import exception_traceback, find_files

from . import __appname__ as APP_NAME
from . import __version__ as APP_VERSION

log = logging.getLogger(__name__)


class LuftdatenPumpe:

    # Live data API URI for luftdaten.info
    uri = 'https://api.luftdaten.info/static/v1/data.json'

    def __init__(self, source=None, filter=None, reverse_geocode=False, progressbar=False, quick_mode=False, dry_run=False):
        self.source = source
        self.reverse_geocode = reverse_geocode
        self.dry_run = dry_run
        self.progressbar = progressbar
        self.filter = filter

        # Quick mode only imports the first few datasets to speed things up.
        self.quick_mode = quick_mode

        # Cache responses from the luftdaten.info API for five minutes.
        # TODO: Make backend configurable.

        # Configure User-Agent string.
        user_agent = APP_NAME + '/' + APP_VERSION

        # Configure cached requests session.
        self.session = CachedSession(
            cache_name='api.luftdaten.info', backend='redis', expire_after=300,
            user_agent=user_agent)

        # Probe Redis for availability.
        try:
            self.session.cache.responses.get('test')
        except redis.exceptions.ConnectionError as ex:
            log.error('Unable to connect to Redis: %s', ex)
            sys.exit(2)

    def get_readings(self):

        if self.source == 'api':
            data = self.request_live_data()

        elif self.source.startswith('file://'):
            path = self.source.replace('file://', '')
            data = self.import_archive(path)

        else:
            raise ValueError('Unknown data source: {}'.format(self.source))

        if data is None:
            raise KeyError('No data selected. Please check connectivity and filter definition.')

        return data

    def get_stations(self):

        if self.source.startswith('postgresql://'):
            return self.make_stations()

        stations = {}
        field_candidates = ['station_id', 'name', 'position', 'location']
        for reading in self.get_readings():

            # Location ID for the reading
            station_id = reading.station.station_id

            # Sensor information from the reading
            sensor_info = Munch({
                'sensor_id': reading.station.sensor_id,
                'sensor_type': reading.station.sensor_type,
            })

            # Acquire additional sensors from reading.
            # This continues with the next loop iteration as
            # location information has already been transferred.
            if station_id in stations:
                station = stations[station_id]
                sensor_id = reading.station.sensor_id
                if not any(map(lambda item: item.sensor_id == sensor_id, station['sensors'])):
                    station['sensors'].append(sensor_info)
                continue

            # New station found: Acquire its information from reading
            station = Munch()
            for field in field_candidates:
                if field in reading.station:
                    station[field] = reading.station[field]

            # Acquire first sensor from reading
            station['sensors'] = [sensor_info]

            # Collect location if not empty
            if station:
                stations[station_id] = station

        results = []
        for key in sorted(stations.keys()):
            station = stations[key]
            results.append(station)

        return results

    def make_stations(self):
        """
        Re-make station information from PostgreSQL database.

        Example::

            {
                "station_id": 28,
                "name": "Ulmer Stra\u00dfe, Wangen, Stuttgart, Baden-W\u00fcrttemberg, DE",
                "position": {
                    "latitude": 48.778,
                    "longitude": 9.236,
                    "altitude": 223.7,
                    "country": "DE",
                    "geohash": "u0wt6pv2qqhz"
                },
                "location": {
                    "place_id": "10504194",
                    "licence": "Data \u00a9 OpenStreetMap contributors, ODbL 1.0. https://osm.org/copyright",
                    "osm_type": "way",
                    "osm_id": "115374184",
                    "lat": "48.777845",
                    "lon": "9.23582396156841",
                    "display_name": "Kulturhaus ARENA, 241, Ulmer Stra\u00dfe, Wangen, Stuttgart, Regierungsbezirk Stuttgart, Baden-W\u00fcrttemberg, 70327, Deutschland",
                    "boundingbox": [
                        "48.7775199",
                        "48.778185",
                        "9.2353783",
                        "9.236272"
                    ],
                    "address": {
                        "country_code": "DE",
                        "country": "Deutschland",
                        "state": "Baden-W\u00fcrttemberg",
                        "state_district": "Regierungsbezirk Stuttgart",
                        "county": "Stuttgart",
                        "postcode": "70327",
                        "city": "Stuttgart",
                        "city_district": "Wangen",
                        "suburb": "Wangen",
                        "road": "Ulmer Stra\u00dfe",
                        "house_number": "241",
                        "neighbourhood": "Wangen"
                    },
                    "address_more": {
                        "building": "Kulturhaus ARENA"
                    }
                },
                "sensors": [
                    {
                        "sensor_id": 658,
                        "sensor_type": "SDS011"
                    },
                    {
                        "sensor_id": 657,
                        "sensor_type": "DHT22"
                    }
                ]
            }
        """

        entries = []

        sql = """
            SELECT *
            FROM ldi_stations, ldi_osmdata
            WHERE
              ldi_stations.station_id = ldi_osmdata.station_id
            ORDER BY
              ldi_stations.station_id
        """

        storage = RDBMSStorage(self.source)
        for station in storage.db.query(sql):
            station = munchify(station)
            entry = {
                "station_id": station.station_id,
                "name": station.name,
                "position": {
                    "latitude": station.latitude,
                    "longitude": station.longitude,
                    "altitude": station.altitude,
                    "country": station.country,
                    "geohash": station.geohash
                }
            }
            entries.append(munchify(entry))

        return entries

    def request_live_data(self):

        log.info('Requesting Live API at {}'.format(self.uri))
        payload = self.session.get(self.uri).content.decode('utf-8')
        data = json.loads(payload)
        #pprint(data)

        timestamp = self.convert_timestamp(data[0]['timestamp'])
        log.info('Timestamp of first record: {}'.format(timestamp))

        iterator = data
        if self.progressbar:
            iterator = tqdm(data)

        for item in iterator:
            try:

                reading = self.make_reading(item)
                if reading is None:
                    continue
                yield reading

            except Exception as ex:
                log.warning('Could not make reading from {}.\n{}'.format(item, exception_traceback()))

    def make_reading(self, item):

        #log.info('item: %s', item)

        # Decode JSON item
        station_id = item['location']['id']
        sensor_id = item['sensor']['id']
        sensor_type = item['sensor']['sensor_type']['name']

        # If there is a filter defined, evaluate it.
        # For specific location|sensor ids, skip further processing.
        if self.filter:
            if 'station' in self.filter:
                if station_id not in self.filter['station']:
                    return
            if 'sensor' in self.filter:
                if sensor_id not in self.filter['sensor']:
                    return

        # Build reading
        reading = Munch(
            station=Munch(),
            data=Munch(),
        )

        # Insert timestamp in appropriate format
        reading.data.time = self.convert_timestamp(item['timestamp'])

        # Insert baseline metadata information
        reading.station.station_id = station_id
        reading.station.sensor_id = sensor_id
        reading.station.sensor_type = sensor_type

        # Collect position information
        del item['location']['id']
        reading.station.position = Munch()
        for key, value in item['location'].items():
            if key in ['latitude', 'longitude', 'altitude']:
                try:
                    value = float(value)
                except:
                    value = None
            reading.station.position[key] = value

        # Collect sensor values
        if 'sensordatavalues' in item:
            for sensor in item['sensordatavalues']:
                name = sensor['value_type']
                value = float(sensor['value'])
                reading.data[name] = value

        # Add more detailed location information
        self.enrich_station(reading.station)

        log.debug('Reading: %s', json.dumps(reading))

        # Debugging
        #break

        return reading

    def enrich_station(self, station):

        # Sanity checks.
        if 'latitude' not in station.position or 'longitude' not in station.position:
            # TODO: Reenable this log message, but just emit once.
            #log.warning('Could not enrich station information for {}'.format(station))
            return

        # Compute geohash.
        station.position.geohash = geohash_encode(station.position.latitude, station.position.longitude)

        # Compute human readable location name.
        if self.reverse_geocode:

            try:

                # Reverse-geocode position.
                station.location = resolve_location(
                    latitude=station.position.latitude,
                    longitude=station.position.longitude,
                    geohash=station.position.geohash,
                    country_code=station.position.country,
                )

                # Improve location information.
                improve_location(station.location)

                # Format address into single label.
                station.name = format_address(station.location)

                try:
                    if station.name.lower() == station.position.country.lower():
                        del station['name']
                except:
                    pass

            except Exception as ex:
                log.error(u'Problem with reverse geocoder for station {}: {}\n{}'.format(station, ex, exception_traceback()))

            if 'name' not in station:
                station.name = u'Station #{}'.format(station.station_id)
                try:
                    station.name += ', ' + station.position.country
                except:
                    pass

    @staticmethod
    def convert_timestamp(timestamp):
        # Mungle timestamp to be formally in ISO 8601 format (UTC).
        if ' ' in timestamp:
            timestamp = timestamp.replace(' ', 'T')
        if '+' not in timestamp:
            timestamp += 'Z'
        return timestamp

    def import_archive(self, path):

        # Optionally append suffix appropriately.
        suffix = '.csv'
        if not path.endswith(suffix):
            path += '/**/*' + suffix

        log.info('Building list of CSV files from {}'.format(path))
        data = list(find_files(path))
        log.info('Processing {} files'.format(len(data)))

        # Optionally add progressbar indicator.
        iterator = data
        if self.progressbar:
            iterator = tqdm(list(data))

        # Process all files.
        for csvpath in iterator:
            readings = self.import_csv(csvpath)
            if readings is None:
                continue

            # Optionally, return first reading only.
            # This is a shortcut-option for churning through the whole data set quickly.
            # This can be used to get whole regional coverage of the data while lacking many details.
            if self.quick_mode:
                readings = list(readings)
                if readings:
                    yield readings[0]
                    continue

            # Process all readings per CSV file.
            for reading in readings:
                if reading is None:
                    continue
                yield reading

    @staticmethod
    def sensor_matches(path, sensors):
        for sensor in sensors:
            if sensor in path:
                return True
        return False

    def import_csv(self, csvpath):

        # Skip files not matching filter pattern (by sensor id).
        if not self.import_archive_filename_accepted(csvpath):
            return

        # When progressbar is enabled, log this to the DEBUG level.
        logger = log.info
        if self.progressbar:
            logger = log.debug
        logger('Reading CSV file {}'.format(csvpath))

        if self.sensor_matches(csvpath, ['ppd42ns', 'sds011', 'pms3003', 'pms5003', 'pms7003', 'hpm']):
            """
            PPD42NS
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;durP1;ratioP1;P2;durP2;ratioP2
            27;PPD42NS;18;;;2015-10-01T19:41:50.266431+00:00;0.62;0;0;0.62;0;0

            SDS011
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;durP1;ratioP1;P2;durP2;ratioP2
            92;SDS011;42;48.800;9.003;2016-07-14T13:02:46.949036+00:00;7.71;;;5.03;;

            PMS3003
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;P2;P0
            409;PMS3003;193;14.382;121.047;2017-12-31T00:00:21;48.00;39.00;

            PMS5003
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;P2;P0
            11390;PMS5003;5751;52.015;5.432;2018-04-20T00:00:25;45.25;31.50;22.50

            PMS7003
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;P2;P0
            5920;PMS7003;2986;51.213;6.808;2017-12-27T00:01:57;5.00;4.00;1.00

            HPM
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;P2
            7096;HPM;3590;50.853;4.342;2017-11-25T00:04:50;22;21
            """
            fieldnames = ['P0', 'P1', 'P2']

        elif self.sensor_matches(csvpath, ['dht22', 'htu21d']):
            """
            DHT22
            sensor_id;sensor_type;location;lat;lon;timestamp;temperature;humidity
            48;DHT22;19;48.722;9.209;2016-08-13T00:00:26.053188+00:00;;

            HTU21D
            sensor_id;sensor_type;location;lat;lon;timestamp;temperature;humidity
            2875;HTU21D;1445;52.600;13.327;2017-11-26T00:00:38;2.80;99.90
            """
            fieldnames = ['temperature', 'humidity']

        elif self.sensor_matches(csvpath, ['bmp180', 'bmp280', 'bme280']):
            """
            BMP180
            sensor_id;sensor_type;location;lat;lon;timestamp;pressure;altitude;pressure_sealevel;temperature
            1464;BMP180;725;48.410;9.934;2017-04-11T00:00:45;96011;;;11.30

            BMP280
            sensor_id;sensor_type;location;lat;lon;timestamp;pressure;altitude;pressure_sealevel;temperature
            2184;BMP280;1098;48.939;8.961;2017-11-26T00:00:35;;;;2.60

            BME280
            sensor_id;sensor_type;location;lat;lon;timestamp;pressure;altitude;pressure_sealevel;temperature;humidity
            1093;BME280;535;48.746;9.211;2017-03-31T00:00:47;96913.02;;;12.75;63.35
            """
            fieldnames = ['temperature', 'humidity', 'pressure', 'altitude', 'pressure_sealevel']

        else:
            log.warning('Skip import of {}. Unknown sensor type'.format(csvpath))
            return

        return self.csv_reader(csvpath, fieldnames)

    def csv_reader(self, csvpath, fieldnames):

        payload = None
        if self.quick_mode:
            try:
                payload = open(csvpath).read(256)
                payload = '\n'.join(payload.split('\n')[:2])
            except:
                pass
        else:
            payload = open(csvpath).read()

        if payload is None:
            log.error('Could not read CSV file %s', csvpath)
            return

        # Read CSV file into tablib's Dataset and cast to dictionary representation.
        imported_data = Dataset().load(payload, format='csv', delimiter=';')
        records = imported_data.dict

        # Iterate all CSV records.
        for record in records:

            item = self.make_item(record)
            try:
                reading = self.make_reading(item)
                if reading is None:
                    continue

                if not self.csvdata_to_reading(record, reading, fieldnames):
                    continue

                yield reading

            except Exception as ex:
                log.warning('Could not make reading from {}.\n{}'.format(item, exception_traceback()))

    def csvdata_to_reading(self, record, reading, fieldnames):

        # Indicator whether we found any data.
        has_data = False

        # Read all the CSV record fields.
        for fieldname in fieldnames:

            # Skip missing keys.
            if fieldname not in record:
                continue

            # Sanitize value.
            value = record[fieldname].strip()

            # Skip empty or non-numeric values.
            if not value or value.lower() == 'nan':
                continue

            # Actually use this reading, casting to float.
            reading.data[fieldname] = float(value)

            # Signal positive dataness.
            has_data = True

        return has_data

    def make_item(self, csvitem):

        log.debug(csvitem)

        item = {
            'timestamp': csvitem['timestamp'],
            'location': {
                'id': int(csvitem['location']),
            },
            'sensor': {
                'id': int(csvitem['sensor_id']),
                'sensor_type': {
                    'name': csvitem['sensor_type'],
                }
            },
        }

        if csvitem['lat'] and csvitem['lon']:
            item['location'].update(latitude=float(csvitem['lat']), longitude=float(csvitem['lon']))

        return item

    def import_archive_filename_accepted(self, csvpath):
        # If there is a filter defined, evaluate it.
        # For specific location|sensor ids, skip further processing.
        if self.filter:

            if 'sensor' in self.filter:
                # Decode station id from filename, e.g. ``2017-01-13_dht22_sensor_318.csv``.
                m = re.match('.+sensor_(\d+)\.csv$', csvpath)
                if m:
                    sensor_id = int(m.group(1))
                    if sensor_id not in self.filter['sensor']:
                        return False

            if 'sensor-type' in self.filter:
                # Decode sensor type from filename, e.g. ``2017-01-13_dht22_sensor_318.csv``.
                m = re.match('.+_(\w+)_sensor_\d+\.csv$', csvpath)
                if m:
                    sensor_type = m.group(1)
                    if sensor_type not in self.filter['sensor-type']:
                        return False

        return True
